@@ -10,12 +10,16 @@ only the local cells that are intersecting it will be updated,
 so will be the corresponding adjacency graph of the complex.
 """
 
+import os
+import string
 from pathlib import Path
 import itertools
 import heapq
 from copy import copy
 from random import random, choices, uniform
+import pickle
 import time
+import multiprocessing
 
 import numpy as np
 from tqdm import trange
@@ -34,7 +38,7 @@ class CellComplex:
     Class of cell complex from planar primitive arrangement.
     """
     def __init__(self, planes, bounds, points=None, initial_bound=None, initial_padding=0.1, additional_planes=None,
-                 build_graph=False):
+                 build_graph=False, quiet=False):
         """
         Init CellComplex.
         Class of cell complex from planar primitive arrangement.
@@ -45,6 +49,8 @@ class CellComplex:
             Plana parameters
         bounds: (n, 2, 3) float
             Corresponding bounding box bounds of the planar primitives
+        points: (n, ) object of float
+            Points grouped into primitives, points[any]: (m, 3)
         initial_bound: None or (2, 3) float
             Initial bound to partition
         build_graph: bool
@@ -52,7 +58,13 @@ class CellComplex:
         additional_planes: None or (n, 4) float
             Additional planes to append to the complex,
             can be missing planes due to occlusion or incapacity of RANSAC
+        quiet: bool
+            Disable logging and progress bar if set True
         """
+        self.quiet = quiet
+        if self.quiet:
+            logger.disabled = True
+
         self.bounds = bounds  # numpy.array over RDF
         self.planes = planes  # numpy.array over RDF
         self.points = points
@@ -289,9 +301,10 @@ class CellComplex:
         extent = bound[1] - bound[0]
         return [bound[0] - extent * padding, bound[1] + extent * padding]
 
-    def _bbox_intersect(self, bound, plane, exhaustive=False, padding=None):
+    def _intersect_bound_plane(self, bound, plane, exhaustive=False, epsilon=10e-5):
         """
-        Bounding box intersection test.
+        Pre-intersection test between query primitive and existing cells,
+        based on AABB and plane parameters.
 
         Parameters
         ----------
@@ -299,10 +312,10 @@ class CellComplex:
             Bound of the query planar primitive
         plane: (4,) float
             Plane parameters
-        padding: None or float
-            Padding for existing cells
         exhaustive: bool
-            Exhaustive partitioning, only for benchmarking.
+            Exhaustive partitioning, only for benchmarking
+        epsilon: float
+            Distance tolerance
 
         Returns
         -------
@@ -310,12 +323,11 @@ class CellComplex:
             Indices of existing cells whose bounds intersect with bounds of the query primitive
             and intersect with the supporting plane of the primitive
         """
+        if exhaustive:
+            return np.arange(len(self.cells_bounds))
 
-        # todo: alpha-shape/convex hull to reduce unnecessary partitioning?
         # each planar primitive partitions only the 3D cells that intersect with it
         cells_bounds = np.array(self.cells_bounds)  # easier array manipulation
-        if padding:
-            bound = self._pad_bound(bound, padding=padding)
         center_targets = np.mean(cells_bounds, axis=1)  # N * 3
         extent_targets = cells_bounds[:, 1, :] - cells_bounds[:, 0, :]  # N * 3
 
@@ -323,13 +335,13 @@ class CellComplex:
             intersection_bound = np.arange(len(self.cells_bounds))
 
         else:
-            # intersection with existing cell AABB
+            # intersection with existing cells' AABB
             center_query = np.mean(bound, axis=0)  # 3,
             center_distance = np.abs(center_query - center_targets)  # N * 3
             extent_query = bound[1] - bound[0]  # 3,
 
             # abs(center_distance) * 2 < (query extent + target extent) for every dimension -> intersection
-            intersection_bound = np.where(np.all(center_distance * 2 < extent_query + extent_targets, axis=1))[0]
+            intersection_bound = np.where(np.all(center_distance * 2 < extent_query + extent_targets + epsilon, axis=1))[0]
 
         # plane-AABB intersection test from extracted intersection_bound only
         # https://gdbooks.gitbooks.io/3dcollisions/content/Chapter2/static_aabb_plane.html
@@ -338,10 +350,7 @@ class CellComplex:
         # compute distance of box center from plane
         distance = np.dot(center_targets[intersection_bound], plane[:3]) + plane[3]
         # intersection between plane and AABB occurs when distance falls within [-radius, +radius] interval
-        intersection_plane = np.where(np.abs(distance) <= radius)[0]
-
-        if exhaustive:
-            return np.arange(len(self.cells_bounds))
+        intersection_plane = np.where(np.abs(distance) <= radius + epsilon)[0]
 
         return intersection_bound[intersection_plane]
 
@@ -383,7 +392,31 @@ class CellComplex:
         """
         return list(self.graph.nodes).index(query)
 
-    def construct(self, exhaustive=False):
+    def _intersect_neighbour(self, kwargs):
+        """
+        Intersection test between partitioned cells and neighbouring cell.
+        Implemented for multi-processing across all neighbours.
+
+        Parameters
+        ----------
+        kwargs: (int, Polyhedron object, Polyhedron object, Polyhedron object)
+            (neighbour index, positive cell, negative cell, neighbouring cell)
+        """
+        n, cell_positive, cell_negative, cell_neighbour = kwargs['n'], kwargs['positive'], kwargs['negative'], kwargs['neighbour']
+
+        interface_positive = cell_positive.intersection(cell_neighbour)
+
+        if interface_positive.dim() == 2:
+            # this neighbour can connect with either or both children
+            self.graph.add_edge(self.index_node + 1, n)
+            interface_negative = cell_negative.intersection(cell_neighbour)
+            if interface_negative.dim() == 2:
+                self.graph.add_edge(self.index_node + 2, n)
+        else:
+            # this neighbour must otherwise connect with the other child
+            self.graph.add_edge(self.index_node + 2, n)
+
+    def construct(self, exhaustive=False, num_workers=0):
         """
         Construct cell complex.
 
@@ -400,14 +433,21 @@ class CellComplex:
         ----------
         exhaustive: bool
             Do exhaustive partitioning if set True
+        num_workers: int
+            Number of workers for multi-processing, disabled if set 0
         """
         logger.info('constructing cell complex')
         tik = time.time()
 
-        for i in trange(len(self.bounds)):  # kinetic for each primitive
+        pool = None
+        if num_workers > 0:
+            pool = multiprocessing.Pool(processes=num_workers)
+
+        pbar = range(len(self.bounds)) if self.quiet else trange(len(self.bounds))
+        for i in pbar:  # kinetic for each primitive
             # bounding box intersection test
             # indices of existing cells with potential intersections
-            indices_cells = self._bbox_intersect(self.bounds[i], self.planes[i], exhaustive)
+            indices_cells = self._intersect_bound_plane(self.bounds[i], self.planes[i], exhaustive)
             assert len(indices_cells), 'intersection failed! check the initial bound'
 
             # half-spaces defined by inequalities
@@ -452,23 +492,18 @@ class CellComplex:
 
                         # adjacency test between both created cells and their neighbours
                         # todo:
-                        #   Avoid 3d-3d intersection if possible. Those unsliced neighbours connect with only one child
-                        #   - reduce computation by half - can be further reduced using vertices/faces instead of
-                        #   polyhedron intersection. Those sliced neighbors connect with both children
+                        #   Avoid 3d-3d intersection if possible. Unsliced neighbours connect with only one child;
+                        #   sliced neighbors connect with both children.
 
-                        for n, cell in enumerate(cells_neighbours):
+                        kwargs = []
+                        for n, cell in zip(neighbours, cells_neighbours):
+                            kwargs.append({'n': n, 'positive': cell_positive, 'negative': cell_negative, 'neighbour': cell})
 
-                            interface_positive = cell_positive.intersection(cell)
-
-                            if interface_positive.dim() == 2:
-                                # this neighbour can connect with either or both children
-                                self.graph.add_edge(self.index_node + 1, list(neighbours)[n])
-                                interface_negative = cell_negative.intersection(cell)
-                                if interface_negative.dim() == 2:
-                                    self.graph.add_edge(self.index_node + 2, list(neighbours)[n])
-                            else:
-                                # this neighbour must otherwise connect with the other child
-                                self.graph.add_edge(self.index_node + 2, list(neighbours)[n])
+                        if pool is None:
+                            for k in kwargs:
+                                self._intersect_neighbour(k)
+                        else:
+                            pool.map(self._intersect_neighbour, kwargs)
 
                     # update cell id
                     self.index_node += 2
@@ -508,8 +543,6 @@ class CellComplex:
             Temp dir to save intermediate visualisation
         """
         if self.constructed:
-            import os
-            import string
             try:
                 import pyglet
             except ImportError:
@@ -517,12 +550,16 @@ class CellComplex:
                 return
             if indices_cells is not None and len(indices_cells) == 0:
                 raise ValueError('no indices provided')
-            temp_filename = ''.join(choices(string.ascii_uppercase + string.digits, k=5)) + '.obj'
-            self.save_obj(filepath=temp_dir + temp_filename, indices_cells=indices_cells, use_mtl=True)
-            scene = trimesh.load_mesh(temp_dir + temp_filename)
+
+            filename_stem = ''.join(choices(string.ascii_uppercase + string.digits, k=5))
+            filename_mesh = filename_stem + '.obj'
+            filename_mtl = filename_stem + '.mtl'
+
+            self.save_obj(filepath=temp_dir + filename_mesh, indices_cells=indices_cells, use_mtl=True)
+            scene = trimesh.load_mesh(temp_dir + filename_mesh)
             scene.show()
-            os.remove(temp_dir + temp_filename)
-            os.remove(temp_dir + 'colours.mtl')
+            os.remove(temp_dir + filename_mesh)
+            os.remove(temp_dir + filename_mtl)
         else:
             raise RuntimeError('cell complex has not been constructed')
 
@@ -600,10 +637,13 @@ class CellComplex:
             points = []
             for cell in self.cells:
                 bbox = cell.bounding_box()
-                samples = [(uniform(bbox[0][0], bbox[1][0]), uniform(bbox[0][1], bbox[1][1]),
-                            uniform(bbox[0][2], bbox[1][2])) for _ in range(num)]
-                mask = [cell.contains(sample) for sample in samples]  # reject sampling
-                points.append(np.array(samples)[np.array(mask)])
+                points_cell = []
+                while len(points_cell) < num:
+                    sample = (uniform(bbox[0][0], bbox[1][0]), uniform(bbox[0][1], bbox[1][1]),
+                              uniform(bbox[0][2], bbox[1][2]))
+                    if cell.contains(sample):
+                        points_cell.append(sample)
+                points.append(points_cell)
             return points
 
         elif location == 'star':
@@ -636,7 +676,7 @@ class CellComplex:
 
         Parameters
         ----------
-        filepath_mesh: str
+        filepath_mesh: str or Path
             Filepath to reference mesh
         engine: str
             Engine to compute predicate, can be 'ray' for ray intersection, or 'distance' for signed distance
@@ -674,13 +714,29 @@ class CellComplex:
         logger.info('number of planes: {}'.format(self.num_planes))
         logger.info('number of cells: {}'.format(self.num_cells))
 
+    def save(self, filepath):
+        """
+        Save the cell complex to a CC file.
+
+        Parameters
+        ----------
+        filepath: str or Path
+            Filepath to save CC file, '.cc' suffix recommended
+        """
+        if self.constructed:
+            # create the dir if not exists
+            with open(filepath, 'wb') as f:
+                pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)
+        else:
+            raise RuntimeError('cell complex has not been constructed')
+
     def save_npy(self, filepath):
         """
         Save the cells to an npy file (deprecated).
 
         Parameters
         ----------
-        filepath: str
+        filepath: str or Path
             Filepath to save npy file
         """
         if self.constructed:
@@ -692,7 +748,7 @@ class CellComplex:
             raise RuntimeError('cell complex has not been constructed')
 
     @staticmethod
-    def _obj_str(cells, use_mtl=False):
+    def _obj_str(cells, use_mtl=False, filename_mtl='colours.mtl'):
         """
         Convert a list of cells into a string of obj format.
 
@@ -702,6 +758,8 @@ class CellComplex:
             Polyhedra cells
         use_mtl: bool
             Use mtl attribute in obj if set True
+        filename_mtl: None or str
+            Material filename
 
         Returns
         -------
@@ -722,7 +780,7 @@ class CellComplex:
         material_str = ''
 
         if use_mtl:
-            scene_str += 'mtllib colours.mtl\n'
+            scene_str += f'mtllib {filename_mtl}\n'
 
         for o in range(len(cells)):
             scene_str += scene_obj[o][0] + '\n'
@@ -742,7 +800,7 @@ class CellComplex:
 
         Parameters
         ----------
-        filepath: str
+        filepath: str or Path
             Filepath to save obj file
         indices_cells: (n,) int
             Indices of cells to save to file
@@ -755,12 +813,12 @@ class CellComplex:
             filepath.parent.mkdir(parents=True, exist_ok=True)
 
             cells = [self.cells[i] for i in indices_cells] if indices_cells is not None else self.cells
-            scene_str, material_str = self._obj_str(cells, use_mtl=use_mtl)
+            scene_str, material_str = self._obj_str(cells, use_mtl=use_mtl, filename_mtl=f'{filepath.stem}.mtl')
 
             with open(filepath, 'w') as f:
                 f.writelines(scene_str)
             if use_mtl:
-                with open(filepath.with_name('colours.mtl'), 'w') as f:
+                with open(filepath.with_name(f'{filepath.stem}.mtl'), 'w') as f:
                     f.writelines(material_str)
         else:
             raise RuntimeError('cell complex has not been constructed')
@@ -771,7 +829,7 @@ class CellComplex:
 
         Parameters
         ----------
-        filepath: str
+        filepath: str or Path
             Filepath to save plm file
         indices_cells: (n,) int
             Indices of cells to save to file
