@@ -28,7 +28,7 @@ class VertexGroup:
     Class for manipulating planar primitives.
     """
 
-    def __init__(self, filepath, process=True):
+    def __init__(self, filepath, process=True, quiet=False):
         """
         Init VertexGroup.
         Class for manipulating planar primitives.
@@ -39,7 +39,12 @@ class VertexGroup:
             Filepath to vertex group file (.vg) or binary vertex group file (.bvg)
         process: bool 
             Immediate processing if set True
+        quiet: bool
+            Disable logging if set True
         """
+        if quiet:
+            logger.disabled = True
+
         if isinstance(filepath, str):
             self.filepath = Path(filepath)
         else:
@@ -491,7 +496,7 @@ class VertexGroup:
 
         Parameters
         ----------
-        filepath: str
+        filepath: str or Path
             Filepath to save txt file
         """
         with open(filepath, 'w') as fout:
@@ -505,7 +510,7 @@ class VertexGroup:
 
         Parameters
         ----------
-        filepath: str
+        filepath: str or Path
             Filepath to save npy file
         """
         logger.info('writing plane parameters into {}'.format(filepath))
@@ -517,7 +522,7 @@ class VertexGroup:
 
         Parameters
         ----------
-        filepath: str
+        filepath: str or Path
             Filepath to save npy file
         """
         logger.info('writing plane bounds into {}'.format(filepath))
@@ -529,20 +534,24 @@ class VertexGroupReference:
     Class of reference vertex group sampled from meshes.
     """
 
-    def __init__(self, filepath, num_samples=10000, process=True):
+    def __init__(self, filepath, num_samples=10000, process=True, quiet=False):
         """
         Init VertexGroupReference.
         Class of reference vertex group sampled from meshes.
 
         Parameters
         ----------
-        filepath: str
+        filepath: str or Path
             Filepath to a mesh
         num_samples: int
             Number of sampled points
         process: bool
             Immediate processing if set True
+        quiet: bool
+            Disable logging if set True
         """
+        if quiet:
+            logger.disabled = True
         import trimesh
 
         self.filepath = filepath
@@ -579,12 +588,12 @@ class VertexGroupReference:
         Start processing mesh data.
         """
         from functools import reduce
+        logger.info('processing {}'.format(self.filepath))
 
         # sample on all faces
         samples, face_indices = self.mesh.sample(count=self.num_samples, return_index=True)  # face_indices match facets
 
         for facet in self.mesh.facets:  # a list of face indices for coplanar adjacent faces
-
             # group corresponding samples by facet
             points = []
             for face_index in facet:
